; Ver 0.01 alpha modified to be accessory decoder from function ver 2.16 - Roco momentary only
; Ver 0.02 Added Lenz support and toggle outputs, protected CV29 fixed bits
; Ver 0.03 Corrected ACK pulse to 6mS

; This is a basic DCC accessory decoder using the 12F630/676 (4 or 8 function)
; It uses the internal 4MHz osc. Execution cycle is 1us.

; It incorporates ideas from D.Probst DCC decoder project
; and Heiko Schroeter's 12F629 mobile 'Z' gauge decoders
; Parts copyright (C) Dean Probst and (C) Heiko Schroeter but released
; under the GNU General Public License

; This version is by Paul Harman mailto:p_harman@hotmail.com and is not fully functional.
; It will be superceded by a working version but is sufficiently functional to test the 
; hardware platform.

;    This program is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.

;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.

;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


; BUGFIXES and ENHANCEMENTS
;
; 1. ACK time reduced to 6mS to be in spec
;
;
; KNOWN BUGS:-
;
;
;
;==========================================================================

 #define ManId .13	; This should be the NMRA assigned ID in CV8, but anything other than 8 or 255
 #define VerNo .4	; The version release number to identify the decoder in CV7

	LIST	p=16F630	; target processor (also possibly work in 16F676 or 12F629))
	__CONFIG  _BODEN_ON & _CP_OFF & _WDT_OFF & _MCLRE_OFF & _INTRC_OSC_NOCLKOUT & _PWRTE_ON
	__idlocs  0xD003

;==========================================================================
;
;       Register Definitions
;
;==========================================================================

W	EQU	H'0000'
F	EQU	H'0001'

;----- Register Files------------------------------------------------------

INDF	EQU	H'0000'
TMR0	EQU	H'0001'
PCL	EQU	H'0002'
STATUS	EQU	H'0003'
FSR	EQU	H'0004'
GPIO	EQU	H'0005'
PORTC	EQU H'0007'	; For 14 pin device compatibility (16F630/676)

PCLATH	EQU	H'000A'
INTCON	EQU	H'000B'
PIR1	EQU	H'000C'

TMR1L	EQU	H'000E'		
TMR1H	EQU	H'000F'		
T1CON	EQU	H'0010'		
CMCON	EQU	H'0019'		
OPTION_REG	EQU	H'0081'
TRISIO	EQU	H'0085'
TRISC	EQU	H'0087'	; For 14 pin device compatibility (16F630/676)
PIE1	EQU	H'008C'
PCON	EQU	H'008E'
OSCCAL	EQU	H'0090'
WPU	EQU	H'0095'
IOCB	EQU	H'0096'
IOC	EQU	H'0096'

VRCON	EQU	H'0099'
EEDATA	EQU	H'009A'	
EEADR	EQU	H'009B'	
EECON1	EQU	H'009C'
EECON2	EQU	H'009D'

ADCON0	EQU H'001F' ; For analogue device compatibility (12F675/16F676)
ADCON1	EQU H'009F' ; For analogue device compatibility (12F675/16F676)
ANSEL	EQU H'0091' ; For analogue device compatibility (12F675/16F676)

;----- STATUS Bits --------------------------------------------------------

IRP	EQU	H'0007'
RP1	EQU	H'0006'
RP0	EQU	H'0005'
NOT_TO	EQU	H'0004'
NOT_PD	EQU	H'0003'
Z	EQU	H'0002'
DC	EQU	H'0001'
C	EQU	H'0000'

;----- GPIO Bits --------------------------------------------------------

GP5	EQU	H'0005'
GPIO5	EQU	H'0005'
GP4	EQU	H'0004'
GPIO4	EQU	H'0004'
GP3	EQU	H'0003'
GPIO3	EQU	H'0003'
GP2	EQU	H'0002'
GPIO2	EQU	H'0002'
GP1	EQU	H'0001'
GPIO1	EQU	H'0001'
GP0	EQU	H'0000'
GPIO0	EQU	H'0000'

;----- INTCON Bits --------------------------------------------------------

GIE	EQU	H'0007'
PEIE	EQU	H'0006'
T0IE	EQU	H'0005'
INTE	EQU	H'0004'
GPIE	EQU	H'0003'
T0IF	EQU	H'0002'
INTF	EQU	H'0001'
GPIF	EQU	H'0000'

;----- PIR1 Bits ----------------------------------------------------------

EEIF	EQU	H'0007'
ADIF	EQU	H'0006'
CMIF	EQU	H'0003'
T1IF	EQU	H'0000'
TMR1IF	EQU	H'0000'

;----- T1CON Bits ---------------------------------------------------------

TMR1GE	EQU	H'0006'
T1CKPS1	EQU	H'0005'
T1CKPS0	EQU	H'0004'
T1OSCEN	EQU	H'0003'
NOT_T1SYNC	EQU	H'0002'
TMR1CS	EQU	H'0001'
TMR1ON	EQU	H'0000'

;----- COMCON Bits --------------------------------------------------------

COUT	EQU	H'0006'
CINV	EQU	H'0004'
CIS	EQU	H'0003'
CM2	EQU	H'0002'
CM1	EQU	H'0001'
CM0	EQU	H'0000'

;----- OPTION Bits --------------------------------------------------------

NOT_GPPU	EQU	H'0007'
INTEDG	EQU	H'0006'
T0CS	EQU	H'0005'
T0SE	EQU	H'0004'
PSA	EQU	H'0003'
PS2	EQU	H'0002'
PS1	EQU	H'0001'
PS0	EQU	H'0000'

;----- PIE1 Bits ----------------------------------------------------------

EEIE	EQU	H'0007'
ADIE	EQU	H'0006'
CMIE	EQU	H'0003'
T1IE	EQU	H'0000'
TMR1IE	EQU	H'0000'

;----- PCON Bits ----------------------------------------------------------

NOT_POR	EQU	H'0001'
NOT_BOD	EQU	H'0000'

;----- OSCCAL Bits --------------------------------------------------------

CAL5	EQU	H'0007'
CAL4	EQU	H'0006'
CAL3	EQU	H'0005'
CAL2	EQU	H'0004'
CAL1	EQU	H'0003'
CAL0	EQU	H'0002'

;----- IOCB Bits --------------------------------------------------------

IOCB5	EQU	H'0005'
IOCB4	EQU	H'0004'
IOCB3	EQU	H'0003'
IOCB2	EQU	H'0002'
IOCB1	EQU	H'0001'
IOCB0	EQU	H'0000'

;----- IOC Bits --------------------------------------------------------

IOC5	EQU	H'0005'
IOC4	EQU	H'0004'
IOC3	EQU	H'0003'
IOC2	EQU	H'0002'
IOC1	EQU	H'0001'
IOC0	EQU	H'0000'

;----- VRCON Bits ---------------------------------------------------------

VREN	EQU	H'0007'
VRR	EQU	H'0005'
VR3	EQU	H'0003'
VR2	EQU	H'0002'
VR1	EQU	H'0001'
VR0	EQU	H'0000'

;----- EECON1 -------------------------------------------------------------

WRERR	EQU	H'0003'
WREN	EQU	H'0002'
WR	EQU	H'0001'
RD	EQU	H'0000'

;==========================================================================
;
;       RAM Definition
;
;==========================================================================

        __MAXRAM H'FF'
        __BADRAM H'06'-H'09', H'0D', H'11'-H'18', H'1A'-H'1F', H'60'-H'7F'
        __BADRAM H'86'-H'89', H'8D', H'8F', H'91'-H'94', H'97'-H'98', H'9E'-H'9F', H'E0'-H'FF'

;==========================================================================
;
;       Configuration Bits
;
;==========================================================================

_CPD_ON		EQU	H'3EFF'
_CPD_OFF	EQU	H'3FFF'
_CP_ON		EQU	H'3F7F'
_CP_OFF		EQU	H'3FFF'
_BODEN_ON	EQU	H'3FFF'
_BODEN_OFF	EQU	H'3FBF'
_MCLRE_ON	EQU	H'3FFF'
_MCLRE_OFF	EQU	H'3FDF'
_PWRTE_OFF	EQU	H'3FFF'
_PWRTE_ON	EQU	H'3FEF'
_WDT_ON		EQU	H'3FFF'
_WDT_OFF	EQU	H'3FF7'
_LP_OSC		EQU	H'3FF8'
_XT_OSC		EQU	H'3FF9'
_HS_OSC		EQU	H'3FFA'
_EC_OSC		EQU	H'3FFB'
_INTRC_OSC_NOCLKOUT	EQU	H'3FFC'
_INTRC_OSC_CLKOUT  	EQU	H'3FFD'
_EXTRC_OSC_NOCLKOUT	EQU	H'3FFE'
_EXTRC_OSC_CLKOUT	EQU	H'3FFF'

;***************** User definable variables in this block ***********************************

#define CV29default B'10000000'	; CV541 default value
									; Bit7=1, Accessory decoder
									; Bit6=0, Decoder address mode
									; Bit5=0, Basic accessory decoder

								
#define Acksecs  .6		; Acknowledge time in mS
AckTime set (Acksecs * D'14') / D'10'	; 1.4 times round the loop approx 1mS @4MHz

;--------------- Do not change anything below this line ----------------------------------------
;-----------------------------------------------------------------------------------------------

; CONSTANTS

#define LokMaus2	.71 	; EEPROM location for LokMaus hundreds
#define Store		.72	; Power down store location for FnSTATE
#define Mode		.33	; CV33 lower 4 bits set toggle top 2 bits set decoder mode
#define ACKbits		.34	; CV34 defines which output pin used for ACK 0=pin 6, 8=pin 11 (5 on 8-pin)
#define Invert		.35	; CV35 like CV99, Output invertion
#define Flicker		.36	; CV36 like CV100 used for flicker map, 0=flicker

; Need to adjust these to move ack to PA2
#define AMask		B'00001000'	; PA0,1,2,4,5 are outputs
#define AMaskComp	B'00001011'	; PA0:1 inputs PA2,4,5 are outputs
#define CMask		B'11010000'	; Set PC5 and PC3:0 output on 14 pin devices (16F684)


; VARIABLES

#define nopreamb	D'19'	; No. of preamble half bits = 20
#define	Seed		.29		; Seed for random number generator


;Bit positions

; Pin connection of 12F629/630 Function decoder
; gp0 = Black track input comparator mode
; gp1 = Red track input comparator mode
; gp2 = Programming acknowledgement ACK pulse
; 
; gp3 = Not used
; gp4 = Function output B
; gp5 = Function output A

; Extra definitions for 14 pin chips
; PC0 = Function output H
; PC1 = Function output G
; PC2 = Function output F
; PC3 = Function output E
; PC4 = Function output D
; PC5 = Function output C

#define gp0	0		; Function F (or Track with comparator)
#define gp1	1		; Function E (or Track with comparator)
#define gp2	2		; Function D
#define gp3	3		; Track (non comparator)
#define gp4	4		; Function B
#define gp5	5		; Function A


;******************************************************
; Bit Positions in CONFIG0, status flags
#define bitrec		0	; What value has just rec. bit
#define CVaccessBit	1	; Two consecutive packets to be send
#define resetflag	6	; Signal reset for service mode
#define smflag		7	; Service Mode Flag

; Bit positions in FLAG
#define FourByte	0	; 1 indicates Direct CV mode in service mode
#define Carry		2	; 0 indicates a borrow in previous subtraction

; Bit positions in CV541
#define AddMode		6	; 1 indicates output address mode

; Bit positions in MODE CV33
#define eight		7	; 1 enables 8-address mode, 0 is 4-address mode
#define Deactivate	6	; 1 enables processing of deactivate commands

;******************************************************
; RAM locations

		cblock	0x20
CONFIG0				; Contains various bits to test
ENDVAL				; Offset to byte check routines
PRECOUNT			; Preamble counter
STATE				; Where are we in the DCC package
DATA1				; First transm. byte
DATA2				; Second trans. byte
DATA3				; Third transm. byte
DATA4				; Fourth transm. byte
DATA5				; Fifth transm. byte
DATA6				; Sixth transm. byte
SAMPLES				; How many samples taken for one or zero
TEMP				; Temporary scratch reg
TEMP1				; ditto
TEMP2				; ditto
TEMP3				; ditto
AddLo				; Part of address to match DATA1
AddHi				; Part of address to match DATA2 high address part
AddOu				; Part of address to match DATA2 output address part
COUNTER    			; 

CVACCESSRATE
CV14
CV29
MYEEDATA
FnSTATE
FnMASK
FLAG				; bit 0=1 four or more bytes to decode, bit 1=1 reverse direction called
TestByte
MODE
ACKBITS
INVERTMAP			; Holds contents of invertmap CV to save on EEPROM reads
OUTPUT				; Values to be applied to output function pins
FlickMASK			; Copy of flicker CV
RANDOM				; Random number updated by flicker routine
		endc
;************************************************************************************************


;************* Here we start ! ******************************************************************************

	org	0
START:
	bsf	STATUS,RP0
	call	0x3FF
	movwf	OSCCAL		;put calibration value into OSCCAL
	bcf	STATUS,RP0

;************* setup sets all what is ness. 
dccsetup:
; Temporarily removed RAM clearance to save space
	movlw	0x20		; bottom of RAM
	movwf	FSR
Next1:
	clrf	INDF
	incf	FSR,F
	movlw	0x60		; end of ram +1
	subwf	FSR,W
	btfss	STATUS,C
	goto	Next1

	movlw	Seed		; Get seed for random number generator
	movwf	RANDOM
	movlw	nopreamb	; Preamble = 20 half bits
	movwf	PRECOUNT
; call master init and read contents of EEPROM into RAM mirrors
	call	ChkPowerUp	; Check EEPROM for consistancy and refresh if required
	call	ReadNMRA	; Load CVs into RAM image buffer for quick access
; Switch off all outputs by loading invert mask into output buffer
	movf	INVERTMAP,W		; Get invertmap	
	movwf	FnSTATE			; Save mask in output buffer
	call	Function_activate	; Activate the outputs
	bsf	STATUS,RP0
	movlw	B'11000000'	; 
	clrf	TRISC		; Set PC5:0 output on 14 pin devices (16F630/676)
	movlw	B'00001011'	; gp2:5 are outputs, GP3 and GP0:1 inputs
	movwf	TRISIO		; 
	bcf		STATUS,RP0
	movlw	0x2			; Leave GP2 as IO and set GP1:0 to comparator in
	movwf	CMCON

; Drop through to starthiC

;**********************************************************************************************
; High Bit Level Section with comparator
starthiC:
	call	Value		;4,5
conthiC:
	btfss	CMCON,COUT	;1	HighLevel ?
	goto	startloC		;2,3	No
	call	Speed_Sub	;3,4
	goto	conthiC		;21,22
;----------------------------------------------------------------------------------------------------
; Low Level Bit Section with comparator
startloC:
	call	Value		;4,5
contloC:
	btfsc	CMCON,COUT	;1
	goto	starthiC		;2,3
	call	Speed_Sub	;3,4
	goto	contloC		;21,22
;----------------------------------------------------------------------------------------------------
; High Bit Level Section with GP3
;starthi:
;	call	Value		;4,5
;conthi:
;	btfss	GPIO,gp3	;1	HighLevel ?
;	goto	startlo		;2,3	No
;	call	Speed_Sub	;3,4
;	goto	conthi		;22,23
;----------------------------------------------------------------------------------------------------
; Low Level Bit Section with GP3
;startlo:
;	call	Value		;4,5
;contlo:
;	btfsc;	GPIO,gp3	;1
;	goto	starthi		;2,3
;	call	Speed_Sub	;3,4
;	goto	contlo		;22,23
;----------------------------------------------------------------------------------------------------
				;Value checks for correct bit and jumps to where we are in DCC package
				;Computed goto has to reside wholy within page 0, so make sure not too
				;much code is added in front.
Value:
	clrf	SAMPLES		; 6
	movf	STATE,W		; 7
	addwf	PCL,F		; 8,9
				;Jump table to routines from where we do a RETURN !
				;cycl. to here ->  cycl for rout.
	goto	waitn		; 10,11
	goto	waitlo
	goto	testlo
; First byte
	goto	bitset		; Half bit
	goto	lastv		; Bit 0
	goto	bitset		; Half bit
	goto	lastv		; Bit 1
	goto	bitset		; Half bit
	goto	lastv		; Bit 2
	goto	bitset		; Half bit
	goto	lastv		; Bit 3
	goto	bitset		; Half bit
	goto	lastv		; Bit 4
	goto	bitset		; Half bit
	goto	lastv		; Bit 5
	goto	bitset		; Half bit
	goto	lastv		; Bit 6
	goto	bitset		; Half bit
	goto	lastx		; Bit 7 so Byte received
; Seperator bits
	goto	end11		; Check first half bit of byte 1-2 seperator
	goto	end12		; Check second half bit of byte 1-2 seperator
	goto	end11		; Check first half bit of byte 2-3 seperator
	goto	end22		; Check second half bit of byte 2-3 seperator
	goto	end31		; Check first half bit of byte 3-4 seperator
	goto	end32		; Check second half bit of byte 3-4 seperator
	goto	end31		; Check first half bit of byte 4-5 seperator
	goto	end42		; Check second half bit of byte 4-5 seperator
	goto	end31		; Check first half bit of byte 5-6 seperator
	goto	end52		; Check second half bit of byte 5-6 seperator
	goto	end61		; Check first half bit of byte 6 terminator
	goto	end62		; Check second half bit of byte 6 terminator
; Decoding will occur after six bytes. If another byte is signalled there will be a frame error

; I guess that nothing should ever get here, but if it does it will check CV8 and return

;******************************************************************************************

ChkPowerUp:
; New, more compact set to factory defaults procedure

; Set the page register to 1 at power on. Page mode will not power cycle between
; setting the page register and programming the CV. A power cycle reset will ensure
; that register and address only modes will always program CV1-4.

; Computed goto so must remain wholy within page 0

	movlw	.1			; Page register default is 1
	movwf	MYEEDATA
	movlw	.0			; Page register uses EEPROM location 0
	call	EEPROM_WRITE

; Check that CV8 has not been changed to force a factory reset
	movlw	0x8		 ; CV8 will contain the manufacturer ID if all is sound
	call	EEPROM_READ
	xorlw	ManId
	btfsc	STATUS,Z
	return			 ; CV 8 is set correctly so return without setting factory defaults

; Set CVs to default values
	movlw	.255		; CV8=255 while resetting to factory default
	movwf	MYEEDATA
	movlw	.8
	call	EEPROM_WRITE

	clrf	TEMP			; TEMP is used as a counter for the table
Next_default:
	Call	GetCVDefault	; Get CV number
	movwF	TEMP1
	incf	TEMP,F
	Call	GetCVDefault	; Get CV contents
	movwf	MYEEDATA
	movf	TEMP1,W
	call	EEPROM_WRITE	; Write CV
	incf	TEMP,F
	movlw	.8
	xorwf	TEMP1,W
	btfsc	STATUS,Z		; Check for CV8, always last
	return					; Done if CV8 done
	goto	Next_default	; Do next CV

GetCVDefault:
	movf	TEMP,W	;
	addwf	PCL,F	;

; Table format is:- 	retlw CV_Number
;						retlw CV_value

	retlw	.1
	retlw	.1		; CV513=1

	retlw	.9
	retlw	.0		; CV521=0

	retlw	.3
	retlw	.0		; CV514=0, output pair 1 constant on

	retlw	.4
	retlw	.0		; CV515=0, output pair 2 constant on

	retlw	.5
	retlw	.0		; CV516=0, output pair 3 constant on

	retlw	.6
	retlw	.0		; CV517=0, output pair 4 constant on

	retlw	.7
	retlw	VerNo	; CV7 = version number

	retlw	.29
	retlw	CV29default	; CV541=NMRA default

	retlw	LokMaus2
	retlw	.0			; Clear LokMaus2 mode hundreds buffer

	retlw	Store
	retlw	.0			; Clear FnSTATE power down storage

	retlw	Mode
	retlw	.0			; CV33 NMRA four pairs no toggle as default

	retlw	ACKbits
	retlw	.0			; CV34 ACK on pin-6 as default

	retlw	Invert
	retlw	b'00000000'	; CV35  No inversions (same as function CV99)

	retlw	Flicker
	retlw	b'11111111'	; CV36  No flickering (same as function CV100)

; CV8 must always be last. CV8=255 while setting to defaults is in progress
	retlw	.8		 ; stamp power up by rewriting manufacturer ID to CV8
	retlw	ManId


;************************************************************************************************************
; Each of the following state routines enters on step 12 and exits with step 22

				; waitn waits for 20 half bits of the preamble
waitn:	
	nop				;12
	nop						;13
	btfss	CONFIG0,bitrec	;14   there are only ones in preamble
	goto	waitn1			;15,16
	decfsz	PRECOUNT,F		;16
	goto	ret4			;17,18
	movlw	nopreamb		;18
	movwf	PRECOUNT		;19
	incf	STATE,F			;20
	return					;21,22

waitn1:
	movlw	nopreamb		;17
	movwf	PRECOUNT		;18

ret4:
	nop						;19
	nop						;20
	return					;21,22

;************************************************************************************************************
				; waitlo waits for low half bit after preamble

waitlo:	
	nop				;12 clear DC timeout when half bit received
	nop						;13
	nop						;14
	nop						;15
	nop						;16
	nop						;17
	btfsc	CONFIG0,bitrec	;18	must be a zero
	goto	ret5			;19,20	might be long preamble so go round again
	incf	STATE,F			;20
ret5:
	return					;21,22

;************************************************************************************************************
				; testlo test second half bit of start bit

testlo:	
	nop				;12
	nop						;13
	nop						;14
	nop						;15
	nop						;16
	nop						;17
	btfsc	CONFIG0,bitrec	;18	must be a zero
	goto	frameerr		;19 incomplete zero bit so no good
	incf	STATE,F			;20
	return					;21,22

;************************************************************************************************************
				; bitset takes first half bit of a bit in the byte

bitset:	
	incf	STATE,F			;12
	bcf	STATUS,C			;13
	btfsc	CONFIG0,bitrec	;14
	bsf	STATUS,C			;15
	rlf	DATA6,F				;16
	nop						;17
	nop						;18
	nop						;19
	nop						;20
	return					;21,22

;************************************************************************************************************
				; lastv checks that first half bit = second half bit
lastv:	
	nop				;12
	nop						;13
	nop						;14
	incf	STATE,F			;15
	btfss	CONFIG0,bitrec	;16
	goto	lastv1			;17,18
	btfss	DATA6,0			;18
	goto	frameerr		;19,20 error if half bits not equal
	nop						;20
	return					;21,22

lastv1:
	btfsc	DATA6,0			;19
	goto	frameerr		;20,21	error if half bits not equal
	return					;21,22
;************************************************************************************************************
				; lastx checks that first half bit = second half bit
				; and sets offset to byte check routine
lastx:	
	nop				;12
	incf	ENDVAL,F		;13
	movf	ENDVAL,W		;14
	addwf	STATE,F			;15
	btfss	CONFIG0,bitrec	;16
	goto	lastx1			;17,18
	btfss	DATA6,0			;18
	goto	frameerr		;19	error if half bits not equal
	nop						;20
	return					;21,22

lastx1:
	btfsc	DATA6,0			;19
	goto	frameerr		;20	error if half bits not equal
	return					;21,22

;************************************************************************************************************
				; end11 end of first byte there must be zero
				; end11 first half bit. end12 second half bit
end11:
	nop				;12
	btfsc	CONFIG0,bitrec	;13
	goto	frameerr		;14,15 first half bit is not zero
	incf	STATE,F			;15
	incf	ENDVAL,F		;16
	nop						;17
	nop						;18
	nop						;19
	nop						;20
	return					;21,22

end12:	
	btfsc	CONFIG0,bitrec	;12
	goto	frameerr		;13,14 second half bit is not zero
	movf	DATA6,W			;14
	movwf	DATA1			;15
	movlw	0x03			;16	point STATE to beginning of jump table
	movwf	STATE			;17
	bcf	FLAG,FourByte		;18
	nop						;19
	nop						;20
	return					;21,22

;************************************************************************************************************
				; end21 end of second byte there must be zero
				; end21 first half bit. end22 second half bit

; end21: is the same as end11:, so use end11:

end22:	
	nop				;12
	btfsc	CONFIG0,bitrec	;13
	goto	frameerr		;14,15 second half bit is not zero
	movf	DATA6,W			;15
	movwf	DATA2			;16
	movlw	0x03			;17	point STATE to beginning of jump table
	movwf	STATE			;18
	nop						;19
	nop						;20
	return					;21,22

;************************************************************************************************************
				; end31 end of third byte there must be a one if last byte, or 0 if more
				; end31 first half bit. end32 second half bit
				; there does not appear to be any testing of the first half bit.
				; Use spare cycles to flag only three bytes.
end31:	
	nop				;12
	incf	STATE,F			;13
	incf	ENDVAL,F		;14
	nop						;15
	nop						;16
	nop						;17
	nop						;18
	nop						;19
	nop						;20
	return					;21,22

end32:	
	nop				;12
	movf	DATA6,W			;13
	movwf	DATA3			;14
	btfsc	CONFIG0,bitrec	;15	if 0 other bytes will follow
	goto	end32x			;16,17 not zero so prepare to decode
	movlw	0x03			;17	point STATE to beginning of jump table
	movwf	STATE			;18
	nop						;19
	nop						;20
	return					;21,22

end32x: 
	clrf	STATE			;18	reset STATE for next preamble
	clrf	ENDVAL			;19
	clrf	DATA4			;20
	clrf	DATA5			;21
	clrf	DATA6			;22
	goto	decode			;23,24	no need here for precise cycle counting, because we decode now
;************************************************************************************************************
				; end41 end of forth byte there must be a one if last byte, or 0 if more
				; end41 first half bit. end42 second half bit
				
				; there does not appear to be any testing of the first half bit	

; end41: is the same as end31:, so use end31:

end42:	
	movf	DATA6,W			;12
	movwf	DATA4			;13
	btfsc	CONFIG0,bitrec	;14	if 0 other bytes will follow
	goto	end42x			;15,16 not zero so prepare to decode
	movlw	0x03			;16	point STATE to beginning of jump table
	movwf	STATE			;17
	bsf	FLAG,FourByte		;18	Indicate that four or more bytes have been decoded
	nop						;19
	nop						;20
	return					;21,22

end42x: 
	clrf	STATE			;17	reset STATE for next preamble
	clrf	ENDVAL			;18
	clrf	DATA5			;19
	clrf	DATA6			;20
	bsf	FLAG,FourByte		;21	Indicate that four bytes have been decoded
	goto	decode			;22,23	no need to count, because we decode now
;************************************************************************************************************
				; end51 end of fifth byte there must be a one if last byte, or 0 if more
				; end51 first half bit. end52 second half bit
				
				; there does not appear to be any testing of the first half bit	

; end51: is the same as end31:, so use end31:

end52:	
	nop				;12
	movf	DATA6,W			;13
	movwf	DATA5			;14
	btfsc	CONFIG0,bitrec	;15	if 0 other bytes will follow
	goto	end52x			;16,17 not zero so prepare to decode
	movlw	0x03			;17	point STATE to beginning of jump table
	movwf	STATE			;18
	nop						;19
	nop						;20
	return					;21,22

end52x: 
	clrf	STATE			;18	reset STATE for next preamble
	clrf	ENDVAL			;19
	clrf	DATA6			;20
	bsf	FLAG,FourByte		;21	Indicate that four bytes have been decoded
	goto	decode			;22,23	no need to count, because we decode now
;************************************************************************************************************
				; end61 end of sixth byte there must be a one since last byte
				; end61 first half bit. end62 second half bit

end61:	
	nop				;12
	btfss	CONFIG0,bitrec	;13
	goto	frameerr		;14,15 first half bit is not one
	incf	STATE,F			;15
	incf	ENDVAL,F		;16
	nop						;17
	nop						;18
	nop						;19
	nop						;20
	return					;21,22

end62:
	nop				;12
	btfss	CONFIG0,bitrec	;13	test correct ending
	goto	frameerr		;14,15 second half bit is not one
	clrf	STATE			;15	reset STATE for next preamble
	clrf	ENDVAL			;16
	goto	decode			;17,18	no need to count, because we decode now
;************************************************************************************************************
				; Frame error in packet
frameerr:
	movlw	nopreamb
	movwf	PRECOUNT
	clrf	STATE
	clrf	ENDVAL
	clrf	DATA1
	clrf	DATA2
	clrf	DATA3
	clrf	DATA4
	clrf	DATA5
	clrf	DATA6
	clrf	SAMPLES

; clear packet status bits in CONFIG0 Byte
	bcf	CONFIG0,bitrec
	bcf	CONFIG0,CVaccessBit
	return

;******************************************************
;************* DECODING *******************************
;******************************************************

decode:
	movf	DATA1,W		; Exclusive or check
	xorwf	DATA2,W
	xorwf	DATA3,W
	xorwf	DATA4,W
	xorwf	DATA5,W
	xorwf	DATA6,W
	btfss	STATUS,Z
	goto	exit_bank1	; error in packet

chkadr:
	movlw	.112
	subwf	DATA1,W
	btfss	STATUS,C	; Less than 112 is a normal short address
	goto	chk_bcast
	movlw	.128
	subwf	DATA1,W
	btfss	STATUS,C	; Greater than 127 is a long address or accessory
	goto	check_sm	; 112-127 is short address but could be service mode
	movlw	B'11000000'
	andwf	DATA1,W
	xorlw	B'11000000'
	btfsc	STATUS,Z
	goto	exit_bank1	; 192 or Greater is long address
;	goto	chknormal	; 128-191 is accessory decoder <= implied goto


;chknormal:
; Check for accessory broadcast address here

	movlw	b'10111111'		; Low part of broadcast address to match DATA1
	xorwf	DATA1,W			; Compare
	btfss	STATUS,Z		; The same?
	goto	baseline		; No, so go on to checking for baseline address
							; Yes, so check DATA2
	movlw	b'10000000'		; High part of broadcast address to match DATA2
	xorwf	DATA2,W			; Compare
	andlw	b'11110000'		; Mask off relevent bits
	btfsc	STATUS,Z		; The same?
	goto	POMchk			; Yes, so check for POM and decode
	goto	baseline		; No, so go on to checking for baseline address

;................................
baseline:
; Set flag if addressed output is in next block
	comf	DATA2,W			; Get low bits of DATA2 complemeted into W
	iorlw	b'11111001'		; Mask off
	addwf	AddOu,W			; Sets carry if addressed output in next block
; Check for normal address
	movf	AddLo,W			; Get low part of address to match DATA1
	btfsc	STATUS,C		; Is this address in next block?
	incf	AddLo,W			; Address is in next block, so check for next block instead
	xorwf	DATA1,W			; Compare
	btfss	STATUS,Z		; The same?
	goto	exit_bank1		; No, so ignore
							; Yes, so check DATA2
	movf	AddHi,W			; Get rest of address to match DATA2
	xorwf	DATA2,W			; Compare
	andlw	b'11110000'		; Mask off relevent bits
	btfss	STATUS,Z		; The same?
	goto	exit_bank1		; No, so ignore
							; Yes, so check for POM

; Need to check here for POM (four byte) commands
POMchk:
	btfsc	FLAG,FourByte		; Four bytes =POM or extended, three bytes=normal
	goto	exit_bank1		; POM so ignore for now
					; Not POM so check for output match

; Temporary debug goto
;	goto	decode_2

;	movf	AddOu,W			; Get rest of address to match DATA2
;	xorwf	DATA2,W			; Compare
;	andlw	b'00000110'		; Mask off relevent bits
;	btfss	STATUS,Z		; The same?
;	goto	exit_bank1		; No, so ignore
;	goto	decode_2		; Yes, so execute command  <= implied goto

;................................

; POM (6 byte):-		DATA1      DATA2      DATA3      DATA4      DATA5      DATA6
;				10AAAAAA 0 1AAACDDD 0 1110CCAA 0 AAAAAAAA 0 DDDDDDDD 0 EEEEEEEE 1

; POM Extended (6 byte):-	DATA1      DATA2      DATA3      DATA4      DATA5      DATA6
; (Not implemented)		10AAAAAA 0 0AAA0AA1 0 1110CCAA 0 AAAAAAAA 0 DDDDDDDD 0 EEEEEEEE 1

; Extended (4 byte):-		DATA1      DATA2      DATA3      DATA4
; (Not implemented)		10AAAAAA 0 0AAA0AA1 0 000DDDDD 0 EEEEEEEE 1

; Accessory (3 byte):-		DATA1      DATA2      DATA3     
;				10AAAAAA 0 1AAACDDD 0 EEEEEEEE 1



;----- Command Decode -------------

decode_2:

;Decode accessory decoder commands here

; Check for Roco turning off mode and jump to Roco if DATA2 bit 3=0
	btfss	DATA2,3			; Test for activate/deactivate
	goto	Roco		; Deactivate so Roco
					; Activate so set output on
; Toggle the outputs depending on CV33

	pageselw	output_control
	rlf	DATA2,W				; Get command and shift up a bit...
	movwf	TEMP
	rlf	TEMP,W				; ... and shift up another bit
	andlw	b'00011100'		; Mask off command
	addwf	PCL,F			; Jump to a 4-instruction command block
	
output_control:
	bsf	FnSTATE,7			; Switch on output A
	btfsc	MODE,0				; Check for toggle mode
	bcf	FnSTATE,6			; Switch off output B
	goto	Output_end

	bsf	FnSTATE,6			; Switch on output B
	btfsc	MODE,0				; Check for toggle mode
	bcf	FnSTATE,7			; Switch off output A
	goto	Output_end

	bsf	FnSTATE,5			; Switch on output C
	btfsc	MODE,1				; Check for toggle mode
	bcf	FnSTATE,4			; Switch off output D
	goto	Output_end

	bsf	FnSTATE,4			; Switch on output D
	btfsc	MODE,1				; Check for toggle mode
	bcf	FnSTATE,5			; Switch off output C
	goto	Output_end

	bsf	FnSTATE,3			; Switch on output E
	btfsc	MODE,2				; Check for toggle mode
	bcf	FnSTATE,2			; Switch off output F
	goto	Output_end

	bsf	FnSTATE,2			; Switch on output F
	btfsc	MODE,2				; Check for toggle mode
	bcf	FnSTATE,3			; Switch off output E
	goto	Output_end

	bsf	FnSTATE,1			; Switch on output G
	btfsc	MODE,3				; Check for toggle mode
	bcf	FnSTATE,0			; Switch off output H
	goto	Output_end

	bsf	FnSTATE,0			; Switch on output H
	btfsc	MODE,3				; Check for toggle mode
	bcf	FnSTATE,1			; Switch off output G
	goto	Output_end			; 

Roco:
	btfss	MODE,Deactivate
	goto	exit_bank1
	pageselw	Roco_output_control
	rlf	DATA2,W				; Get command and shift up a bit
	andlw	b'00001110'		; Mask off command
	addwf	PCL,F			; Jump to a 2-instruction command block
	
Roco_output_control:
	bcf	FnSTATE,7			; Switch off output A
	goto	Output_end
	bcf	FnSTATE,6			; Switch off output B
	goto	Output_end
	bcf	FnSTATE,5			; Switch off output C
	goto	Output_end
	bcf	FnSTATE,4			; Switch off output D
	goto	Output_end
	bcf	FnSTATE,3			; Switch off output E
	goto	Output_end
	bcf	FnSTATE,2			; Switch off output F
	goto	Output_end
	bcf	FnSTATE,1			; Switch off output G
	goto	Output_end
	bcf	FnSTATE,0			; Switch off output H
;	goto	Output_end			; <= implied


Output_end:
	pageselw	Value
	call	Function_activate		; Put FnSTATE onto the outputs
	goto	exit_bank1

;*****************************************************************************************************************


;*****************************************************************************************************************

EEPROM_WRITE:
; Address in W, DATA in MYEEDATA
	bsf	STATUS,RP0
;	movwf	EEADR		<- superfluous if reading first
	call	EEPROM_READ	; get previously stored value
	bcf	STATUS,RP0
	xorwf	MYEEDATA,W	; Compare with new value
	btfsc	STATUS,Z	; If zero no need to write
	return				; No need to write so return
	movf	MYEEDATA,W	; Load value to be written
	bsf	STATUS,RP0
	movwf	EEDATA
	bsf	EECON1,WREN
	movlw	0x55
	movwf	EECON2
	movlw	0xAA
	movwf	EECON2
	bsf	EECON1,WR		; Instigate write sequence
wait_ee_write:
	btfsc	EECON1,WR	; Wait for write to complete
	goto	wait_ee_write
	bcf	STATUS,RP0
	return

EEPROM_READ:
; Address in W
; return data in W
	bsf	STATUS,RP0
	movwf	EEADR
	bsf	EECON1,RD
	movf	EEDATA,W
	bcf	STATUS,RP0
	return

;************************************************************************************************************
ReadNMRA:
; Need to read CV541 first in order to interpret the other CVs

	movlw	.29			; CV541 = Configuration bits
	call	EEPROM_READ
	movwf	CV29

; CV513 can be the decoder address if CV541 bit 6 is clear, or output address if CV541 bit 6 is set
; CV513 and CV521 both =0 is invalid since this is the broadcast address
	movlw	.1			; CV513 = basic address
	call	EEPROM_READ
	addlw	.255		; Subtract 1
	bsf	FLAG,Carry		; Copy carry to flag overflow
	btfss STATUS,C
	bcf	FLAG,Carry
	movwf	AddLo		; Store in low address
	andlw	b'00000011'	; Mask off unwanted bits
	movwf	AddOu		; And store in output address

	movlw	.9			; CV521 = high address
	call	EEPROM_READ
	btfss FLAG,Carry	; Check for underflow on previous subtraction
	addlw	.255		; Subtract 1 if a borrow from previous subtraction
	andlw	b'00000111'	; Mask off unwanted bits
	xorlw	b'00001111' ; 1's complement
	movwf	AddHi		; And store in high address

	btfss	CV29,AddMode	; Decoder address mode or output address mode?
	goto	NoShift			; Bit clear so decoder address mode
							; Set so output address mode

; AddLo looks like this:-		AAAAAAxx

	rrf	AddLo,F			; Shift down
	rrf	AddLo,F			; And shift again, now xxAAAAAA

NoShift:
	bsf AddLo,7			; Now 1xAAAAAA
	bcf AddLo,6			; Now 10AAAAAA

; AddLo now looks like this:-	10AAAAAA to match DATA1 in received packet

; AddOu looks like this:-		000000AA

	bcf	STATUS,C		; Clear carry ready for shift
	rlf	AddOu,F			; Shift

; AddOu now looks like this:-	00000AA0 to match DATA2 low nibble

; AddOu will always be zero in decoder address mode, so clear if required

	btfss	CV29,AddMode	; Decoder address mode or output address mode?
	clrf	AddOu			; Clear AddOu since it is always zero

; AddHi looks like this:-		00001AAA

	swapf	AddHi,F		; Swap the nibbles

; AddHi now looks like:-		1AAA0000 to match DATA2 high nibble

	movlw	Mode		; CV545 (CV33) = toggle mode
	call	EEPROM_READ
	movwf	MODE

	movlw	ACKbits		; CV546 (CV34) = ACK pin
	call	EEPROM_READ
	movwf	ACKBITS

	movlw	Invert		; CV547 (CV35) = output invert map
	call	EEPROM_READ
	movwf	INVERTMAP

	movlw	Flicker		; CV548 (CV36)= flicker map
	call	EEPROM_READ
	movwf	FlickMASK

	goto	exit_bank1	; start sampling

;************* SERVICE MODE *******************************************

; Direct CV (4 byte):-		DATA1      DATA2      DATA3      DATA4
;							1110CCAA 0 AAAAAAAA 0 DDDDDDDD 0 EEEEEEEE 1

; Legacy modes (3 byte):-	DATA1      DATA2      DATA3     
;							1110CAAA 0 DDDDDDDD 0 EEEEEEEE 1



chk_bcast:
; Check for general broadcast address here

	movf	DATA2,W		; Set ZERO Flag if a broadcast address
	btfsc	STATUS,Z
	goto	decode_1	; zero is valid Broadcast address
	goto	exit_bank1	; Not broadcast so ignore

check_sm:
	btfss	CONFIG0,resetflag	; check for SM, reset packet has to come first
	goto	exit_bank1		; No reset packet so must be short address 112-127

	btfss	CONFIG0,smflag	; Second SM packet
	goto	set_sm_flag

; Direct mode is four bytes, but other modes are only three (inc check byte)
; Use FLAG,FourByte=1 to indicate Direct CV mode

	btfss	FLAG,FourByte	; Direct CV Mode if set, else not enough bytes so legacy
	goto	ad_mode_write	; Go to legacy address only/register/paged mode

; Direct CV mode. This bit manipulates the data to be the same as POM (short address).
dm_mode_write:			; Drop through to Direct CV mode
	movf	DATA3,W		; Data in Byte 3
	movwf	DATA4		; now CV contents in DATA 4
	movf	DATA2,W		; CV number in Byte 2, 0=CV1
	movwf	DATA3		; Now most of CV number (all the useful bit) in DATA3
	movf	DATA1,W		; Command in Byte 1
	movwf	DATA2		; Command now in DATA2
	goto	DoCV		; Carry on as if POM

; Address only/register/paged mode. This part manipulates the data to be the same as POM.
; Losing support for modes other than Direct CV will save about 50 program words! 

ad_mode_write:
	movf	DATA2,W		; Data in byte 2
	movwf	DATA4		; Now CV contents in DATA4
	movlw	B'00001100'	; Write command for POM/Direct CV ready 
	btfss	DATA1,3		; Test that write mode is what is required
	movlw	B'00000100'	; Verify command for POM/Direct CV if not write mode
	movwf	DATA2		; Command Verify or Write now in DATA2
	movf	DATA1,W		; Get register address from command byte 
	andlw	b'00000111'	; Mask off CV address (lower 3 bits)
	movwf	DATA3		; Put CV address in DATA3 (0=CV1)
	btfsc	DATA3,2		; Test for CV1-4 or CV5-8
	goto	Not_page	; Not in the paged bank, so sort out CV541 and page register
; This bit adds on the page offset
	movlw	.0			; Lets see what is in the page register, EEPROM location 0
	call	EEPROM_READ	; Read the page register
	movwf	TEMP
	decf	TEMP,F		; Take off one from page register to get multiplier
	movlw	b'11100000'	; Check for out of range page register
	andwf	TEMP,W		; Zero result for valid page value
	btfss	STATUS,Z	; Carry on if page in range
	goto	sm_error	; Out of range so reset and don't program
	bcf		STATUS,C	; Clear carry ready for multiply
	rlf		TEMP,F		; times 2
	bcf		STATUS,C	; Clear carry ready for multiply
	rlf		TEMP,W		; Times 2 again (2x2=4). Page multiplier now in top 6 bits of W
	iorwf	DATA3,F		; Data 3 should now be 0PPPPPAA, or CV address, 0=CV1
	goto	DoCV		; Carry on as if POM

Not_page:	; W contains CV address 100-111, CV541,page register,CV7 and CV8
	btfsc	DATA3,1		; Test for CV541/page register or CV7-8
	goto	DoCV		; 11X, CV7 or 8, Carry on as if POM
	btfsc	DATA3,0		; Test for CV541 or page register
	goto	Set_page	; 101, set page register
	movlw	.28			; CV541 POM address = 28
	movwf	DATA3		
	goto	DoCV		; Carry on as if POM

Set_page
; Bit of a fudge here, set page register manually then refresh CV127 as a dummy.
; Write of CV127 will generate all the acknowledgement and stuff

	movf	DATA4,W		; Get value that needs to go into the page register
	movwf	MYEEDATA	; Load EEPROM with page value
	movlw	0			; EEPROM 0 is page register
	call	EEPROM_WRITE	; write page register
	movlw	.126
	movlw	DATA3		; Set DATA3 to CV127 address
	movlw	.127
	call	EEPROM_READ	; Read CV127 so it can be rewritten!
	movwf	DATA4		; Put CV127 contents into DATA4
	goto	DoCV		; Carry on and write CV127 as if POM. Page register already written!

;******************CV Access group*******************************

; Routine for 'Operations mode' or 'Programming on the main' and 'Direct CV' modes
; all modes converted to be compatible with short address POM for DoCV routine

;POM (short address) :-	DATA1      DATA2      DATA3      DATA4      DATA5
;						0aaaaaaa 0 1110CCAA 0 AAAAAAAA 0 DDDDDDDD 0 EEEEEEEE 1
;
;POM (long address) :-	DATA1      DATA2      DATA3      DATA4      DATA5      DATA6
;						11aaaaaa 0 aaaaaaaa 0 1110CCAA 0 AAAAAAAA 0 DDDDDDDD 0 EEEEEEEE 1

; The CV address is 1 less than the CV number so address 00 00000000 = CV1
; CC Field, 11=Write, 01=Verify, 10=Bit manipulation (not yet tested)
; a=decoder address, A=CV address, D=CV contents, E=checksum

; For bit manipulation the DDDDDDDD field is broken up into 111KDBBB where D is the
; bit value, BBB is the bit position within the CV, K=1 means write, K=0 means verify


CVaccess:
	btfsc	CONFIG0,CVaccessBit	; second packet received ?
	goto	DoCV
	bsf	CONFIG0,CVaccessBit
	goto	exit_bank1	
DoCV:
	bcf	CONFIG0,CVaccessBit
; LokMaus2 support mode here getting value from Lokmaus location in EEPROM that was written to CV7
;
; Process is to write hundreds columns to CV7 then use tens and units to write CV
;
;            CV7 = Hh                   <- First write
;                 /  \
;                /    \
;       CV No = HTU    htu = CV value
;                ||     ||
;        CV No = TU     tu = Value      <- Second write
;
; This will result in value htu being put in CV HTU
;
; Bit values in Lokmaus EEPROM location indicate the following:-
;	0 (1) = 1,  Add 100 to CV contents
;	1 (2) = 1,  Add 200 to CV contents if bit 3 = 0 (02 or 22 programmed to CV7)
;	2 (4) = 1,  Add 200 to CV contents if bit 3 = 1 (12 programmed to CV7)
;	3 (8) = 1,  Add 100 to CV number
;	4 (16) = 1, Add 200 to CV number
;
; Subtracting 2 from Lokmaus EEPROM contents if bit 3 = 1 simplifies things as follows:-
;	0 (1) = 1,  Add 100 to CV contents
;	1 (2) = 1,  Add 200 to CV contents
;	2 (4) = 1,  ignore
;	3 (8) = 1,  Add 100 to CV number
;	4 (16) = 1, Add 200 to CV number
;
; Get CV7 value from EEPROM, and subtract 2 if required
	movlw	LokMaus2		; Read CV7 saved value
	call	EEPROM_READ		; Get current contents
	movwf	TEMP			; Save CV7 contents into TEMP
	movlw	.2				; Get 2 ready in case it neads to be subtracted
	btfsc	TEMP,3			; Test for 1x
	subwf	TEMP,F			; Take away 2 to leave bit 1 free of tens column
; Add digits from CV 7 to CV number and value
	movf	DATA4,W			; Get CV value
	btfsc	TEMP,0			; Test for x1
	addlw	.100			; Add 100 to CV value
	btfsc	TEMP,1			; Test for x2
	addlw	.200			; Add 200 to CV value
	movwf	DATA4			; Save modified value
	movf	DATA3,W			; Get CV number
	btfsc	TEMP,3			; Test for 1x
	addlw	.100			; Add 100 to CV number
; Only CV1-127 currently supported, so 2xx will be out of range.
;	btfsc	TEMP,4			; Test for 2x
;	addlw	.200			; Add 200 to CV number
	movwf	DATA3			; Save modified CV number
	clrf	MYEEDATA		; Clear Lokmaus location
	movlw	LokMaus2		; LokMaus2 EEPROM = CV7 saved location
	call	EEPROM_WRITE	; Write LokMaus2

;Test_CV:
	movf	DATA2,W			; Load top two bits
	andlw	B'00000011'		; Mask off from instruction
	btfss	STATUS,Z		; check that they are both zero
	goto	exit_bank1		; CV is above CV256 if non zero
	movf	DATA3,W			; load lowest 8 bits of CV address
	movwf	TEMP
	incfsz	TEMP,F			; Address 0 is CV1 so increase to get CV and EEPROM number
;	btfsc	TEMP,7			; Valid CV for this decoder is CV1-127
;	goto	exit_bank1		; CV is above 127
	movf	DATA2,W
	andlw	B'00001100'
	xorlw	B'00001100'		; CC=11=Write 
	btfsc	STATUS,Z
	goto	cv_prog
	xorlw	B'00001000'		; CC=01=Verify 
	btfsc	STATUS,Z
	goto	cv_verify
	xorlw	B'00001100'		; CC=10=Bit manipulation 
	btfss	STATUS,Z
	goto	exit_bank1		; CC=00=no valid CV command

; Convert bit manipulation into standard write/verify commands
	movlw	B'00001100'	; Write command 
	btfss	DATA4,4		; Test that write mode is what is required
	movlw	B'00000100'	; Verify command if not
	movwf	DATA2		; Command Verify or Write now in DATA2

; Convert bit manipulation data field into standard CV write/verify format
	movf	DATA4,W		; Get the bit position number from the data
	andlw	b'00000111'	; Mask off all but the bit position
	movwf	TEMP1		; Store in counter
	movf	TEMP,W		; Get CV number
	call	EEPROM_READ	; Get current contents of CV
	movwf	TEMP3		; Put the current contents into TEMP3
	clrf	TEMP2
	bsf		STATUS,C	; Set a bit in the carry ready to rotate
	incf	TEMP1,F		; Zero is one bit, not zero bits!!!
Multiply:
	rlf		TEMP2,F		; Shift a bit up
	decfsz	TEMP1,F		; Is that enough shifting?
	goto	Multiply	; No
	movf	TEMP2,W		; TEMP2 now contains one bit in the correct position
	iorwf	TEMP3,F		; Set the bit in TEMP3
	xorlw	b'11111111'	; Complement the bit position (a zero now in the bit position)
	btfss	DATA4,3		; Test the bit
	andwf	TEMP3,F		; Clear the bit if it should be zero
	movf	TEMP3,W		; Get the manipulated data field
	movwf	DATA4		; And put it in DATA4 to be processed as a normal 8 bit CV value

cv_prog:
; Check decoder lock here before writing
; CV number in TEMP, CV value in DATA4
	movlw	.1				; Check for CV1
	xorwf	TEMP,W			; Compare CV number with 1
	btfsc	STATUS,Z		; Not CV1 so check next
	goto	Write_OK		; CV1 so write even if locked
	movlw	.15				; Check for CV15
	xorwf	TEMP,W			; Compare CV number with 15
	btfsc	STATUS,Z		; Not CV15 so check next
	goto	Write_OK		; CV15 so write even if locked
	movlw	.15				; Get CV15 for comparison
	call	EEPROM_READ		; CV15 now in W
	movwf	TEMP2			; Store contents of CV15
	xorlw	.255			; Check for 'broadcast' value
	btfsc	STATUS,Z		; CV15 not 255 so check next
	goto	Write_OK		; CV15 255 so write even if locked
	movlw	.16				; Get CV16 to compare
	call	EEPROM_READ		; CV16 now in W
	xorwf	TEMP2, W		; Compare CV15 and CV16 contents
	btfss	STATUS,Z		; CV15 and CV16 are equal, so write
	goto	exit_bank1		; Not equal so exit without writing
Write_OK:
	movf	DATA4,W			; value in DATA4
	movwf	MYEEDATA
	movf	TEMP,W			; CV number held in TEMP
; Check for CV7 write for LokMaus2 mode
	xorlw	.7
	movlw	LokMaus2		; Get CV7 saved location ready
	btfsc	STATUS,Z		; Is it CV 7?
	movwf	TEMP			; Yes, so put LokMaus2 location in TEMP
; Check for CV29 write, must not allow illegal bits to be set
	movlw	b'10000000'		;
	xorwf	MYEEDATA		; Toggle top bit
	movf	TEMP,W			; CV number held in TEMP
	xorlw	.29
	movlw	b'01000000'		; Get mask ready 
	btfsc	STATUS,Z		; Is it CV 29?
	andwf	MYEEDATA,F		; Yes, so clear lower bits and set top bit inverted
	movlw	b'10000000'		;
	xorwf	MYEEDATA		; Toggle top bit back
; Write the CV
	movf	TEMP,W			; Get CV number from TEMP
	call	EEPROM_WRITE	; write CV
	movf	TEMP,W			; CV number held in TEMP

cv_verify:
	call	EEPROM_READ		; Read back stored CV contents into W
	xorwf	DATA4,W			; Compare read value with written value
	btfss	STATUS,Z		; equal than ACK
	goto	exit_bank1		; not equal exit without ACK
;---------------------------------------------------------------

acknowledge:

	movlw	0xFF
	movwf	GPIO	; Turn outputs A,B and D on at least
	movwf	PORTC	; Turn on the rest as well
	clrf	TEMP1
	movlw	AckTime		; acktime in ms can be set in *.h file
	movwf	TEMP
loop_ack1:
	decfsz	TEMP1,F
	goto	loop_ack1
	decfsz	TEMP,F
	goto	loop_ack1

	clrf	GPIO	; turn off
	clrf	PORTC
exit_cv:
	bcf	CONFIG0,smflag	; clear SM mode in case
	bcf	CONFIG0,resetflag
	goto	ReadNMRA

set_sm_flag:
	bsf	CONFIG0,smflag
	goto	exit_bank1

sm_error:
	bcf	CONFIG0,resetflag
	bcf	CONFIG0,smflag
	goto	exit_bank1

; This bit is to decode broadcast reset packets when in service mode
decode_1:
	movf	DATA2,W		; Command decoding jump table
	andlw	0xE0		; Isolate command

	xorlw	.0			; Consist (000.....)
	btfsc	STATUS,Z
	goto	ConsistGroup

	bcf	CONFIG0,resetflag
	bcf	CONFIG0,smflag

	goto	exit_bank1	; Only need to do decoder reset in consist group so ignore the rest



Function_activate:
; FnSTATE now contains the the active state of outputs A to H.
; Invert according to CV
	movf	INVERTMAP,W	; Get invert map
	xorwf	FnSTATE, W	; Invert required bits of function state and leave in W
	movwf	PORTC		; Put outputs C-H on Port C, C=PC5,D=PC4,E=PC3,F=PC2,G=PC1,H=PC0
	movwf	TEMP		; Store state ready for shifting
	rrf		TEMP, F		; Shift outputs A-H down one (H into carry)
	rrf		TEMP, W		; Shift outputs A-H down another one (G into carry)
	movwf	GPIO		; stick outputs A-F on GPIO, A=GP5,B=GP4,C=GP2,E=GP1,F=GP0
	return

;***********************************************************************************************************************
				; Consist Group
; Decoder reset apears as a consist group instruction to the broadcast address
; I think that this bit may contain errors!

ConsistGroup:
	bcf	CONFIG0,CVaccessBit	; clear CVaccess flag
	movf	DATA2,W
	xorlw	.0			; Service mode decoder reset instruction 00000000 
	btfsc	STATUS,Z
	goto	soft_reset	;
	xorlw	.1			; service mode hard reset instruction 00000001
	btfsc	STATUS,Z
	goto	hard_reset	; Set CV29 to default and clear CV19.

; Looks like the rest of this is not relevent to accessory decoder so skipping


; Check here for setting/reseting of CV29 bit 5 using SM inst. 0000101x
;	andlw	B'11111110'
;	xorlw	B'00001010'	; Service mode set advanced addressing instruction 0000101x
;	btfsc	STATUS,Z
;	goto	long_adr_set

; Start checking for consist instructions.
; 0001001x where x is the direction of travel relative to the consist (CV19 bit 7)
;	movf	DATA2,W
;	andlw	B'11111110'
;	xorlw	B'00010010'
;	btfsc	STATUS,Z
;	goto	cst_ctrl_set
	goto	exit_bank1

hard_reset:
; Perform NMRA service mode Hard Reset
	clrf	MYEEDATA	; no consist
	movlw	.19			; get EEPROM location of CV19
	call	EEPROM_WRITE

	movlw	CV29default	; CV29	
	movwf	MYEEDATA
	movlw	.29		; get EEPROM location of CV29
	call	EEPROM_WRITE

; reset CV31 and CV32 here as well if implemented.

soft_reset:
; Service mode reset packet received, so prepare for service mode
	bsf	CONFIG0,resetflag	; set resetflag for SM mode detection
; Decoder will now respond to address 112-127 as service mode packets
; Make sure that all function outputs are inactive
	clrf	FnSTATE			; Turn all functions off
	call	Function_activate
	goto	ReadNMRA

; Set consist address and direction in CV19
;cst_ctrl_set:
;	bsf		DATA3,7
;	btfss	DATA2,0
;	bcf		DATA3,7
;	movf	DATA3,W
;	movwf	MYEEDATA
;	movlw	.19
;	call	EEPROM_WRITE
;	goto	exit_bank1

; Set/reset bit in CV29 for long addressing 
long_adr_set:
;	movlw	.29
;	call	EEPROM_READ	; Get CV29 contents
;	movwf	MYEEDATA	; Put in buffer
;	bsf		MYEEDATA,5	; Set bit
;	btfss	DATA2,0		; Check bit
;	bcf		MYEEDATA,5	; Clear bit if it should not be set
;	movlw	.29
;	call	EEPROM_WRITE; Rewrite CV29
;	goto	exit_bank1	<- implied goto

;***********************************************************************************************************************
exit_bank1:
	clrf	TEMP
	clrf	DATA1
	clrf	DATA2
	clrf	DATA3
	clrf	DATA4
	clrf	DATA5
	clrf	DATA6
	clrf	ENDVAL
	return
;***********************************************************************************************************************

; Timing is critical, enters on step 5 and should exit with last step of return step 20
Speed_Sub
	incf	SAMPLES,F	;5
	bcf	CONFIG0,bitrec	;6
	movlw	0xFD		;7
	addwf	SAMPLES,W	;8
	btfss	STATUS,C	;9
	bsf	CONFIG0,bitrec	;10
	decfsz	COUNTER,F	;11
	Goto	Do_Flicker	;12,13	Update random number generator
	movf	FnSTATE,W	;13	Get function state
	btfss	RANDOM,7	;14	Test random bit
	andwf	FlickMASK,W	;15	Bit clear, so switch off flicker outputs
	xorwf	INVERTMAP,W	;16 	Invert required bits of function state and leave in W
	movwf	PORTC		;17 	Put outputs C-H on Port C, C=PC5,(D=PC4),E=PC3,F=PC2,G=PC1,H=PC0
	nop			;18	No analogue support for accessories	
	return				;19,20 timeout not expired, so carry on receiving bits

Do_Flicker:
; Randomisation routine
	movlw	01DH		;14 Load prime number
	clrc				;15 Clear carry ready for shift
	rlf		RANDOM,F	;16
	btfsc	STATUS,C	;17
	xorwf	RANDOM,F	;18
; Random bit supplied in RANDOM,7 
	return				;19,20


	org	0x2108	; Make sure that CV8=255 to force load of EEPROM to defaults on first boot
	DW	.255	; CV8=255

	END
